#ifndef PROBE_GROUP2_HXX 
#define PROBE_GROUP2_HXX

#ifdef PROBE_GROUP_HXX
#error You cant include probe_group.hxx AND probe_group2.hxx
#endif

#include <vector>
#include <string>
#include <iostream>
#include <list>
#include <set>

#include "arbdb.h"
#include "arbdbt.h"

#ifdef NDEBUG
#define pg_assert(bed)
#else
#define pg_assert(bed) do { if ((bed)==false) { cerr << "Assertion '" << #bed << "' failed in " << __LINE__ << "\n"; pg_guard(); exit(1); } } while(0)
#endif
    

using namespace std;

extern GBDATA *gb_main;

#ifndef NDEBUG
void pg_guard(void);
#endif

#ifdef SUN5
template <class T, class Allocator _RWSTD_COMPLEX_DEFAULT(allocator<T>) >
class checked_vector : public vector<T,Allocator>
#else
template <class T>
class checked_vector : public vector<T>
#endif
{
public:
  T& at(size_t n) { pg_assert(n>=0 && n<size()); return (*this)[n]; }
  const T& at(size_t n) const { pg_assert(n>=0 && n<size()); return (*this)[n]; }
};


typedef checked_vector<string> string_container;
typedef set<int,less<int> > setA;
typedef checked_vector<const GBT_TREE*> PG_tree_nodes;

class Specie {

public:
  GBDATA *gbp;

  //Zeiger auf platz im Baum
  const GBT_TREE* tree_place;

  Specie( GBDATA *gbp1){
    gbp=gbp1;
    pg_assert(gbp);
    tree_place = 0;
  }
  Specie(const GBT_TREE* p){
    tree_place=p;
    gbp=p->gb_node;
    pg_assert(gbp);
    pg_assert(tree_place);
  }

  const char *get_aligm(char *use, GB_ERROR& error) const;
  const char *get_name(GB_ERROR& error) const; 
 
  bool operator!=(const Specie& other) const { return gbp!=other.gbp; } 
  bool operator==(const Specie& other) const { return gbp==other.gbp; } 

  Specie(const Specie& other);
  Specie& operator=(const Specie& other);
    
}; 



class NextSond{

  string zustand;
 
  int counter_first;
  int counter_second;
  int nukleotid_first;// 1-A, 2-C, 3-G, 4-U
  int nukleotid_second; 
  int current_length;
  int max_length;

public:

  string get_zustand(){return zustand;}
  void set_zustand(string s){zustand=s;}

  void init(){ zustand="AAAAAAAAAAAAAAA";max_length=20;nukleotid_first=1;nukleotid_second=1;
               current_length=15;counter_first=current_length-1;counter_second=current_length;}
  const string& getnext_Sonde();
  void print_sonde(){printf("%s\n",zustand.c_str());printf("current_length=%d\n",current_length);}

}; 




//Bestimmungen fuer Suchfunction
class PG_define{
public:
  int probelength;//Groesse der Probe
  int cliptresult;//Maximale Anzahl von Sonden
  double mintemp;//Minimale Temperatur
  double maxtemp;//Maximale Temperatur
  int mishit;//Max non group hits
  double mintargets ;//Min group hits (%) /100
  double mingc;  //G +C content
  double maxgc;//G + C content 
  int minpos;//ECOLI-Position
  int maxpos;//ECOLI-Position  
  double maxbonds; //=4
  double bondval[16];
  double dtedge;
  double dt;
  double split;

  void set_probelength(int len){ probelength=len;}
  void set_cliptresult(int result){cliptresult=result;}
  void set_mintemp(double mtemp){mintemp=mtemp;}
  void set_maxtemp(double matemp){maxtemp=matemp;}
  void set_mishit(int hit){mishit=hit;}
  void set_mintargets(double targets){mintargets=targets;}
  void set_mingc(double mgc){mingc=mgc;}
  void set_maxgc(double magc){maxgc=magc;}
  void set_minpos(int mpos){minpos=mpos;}
  void set_maxpos(int mapos){maxpos=mapos;}
  void set_dtedge(double in){dtedge=in;}
  void set_dt(double in){dt=in;}
  void set_split(double in){split=in;}
  void set_bondval(double in,int count){bondval[count]=in;}
  void set_maxbonds(double n){maxbonds=n;}
};
typedef checked_vector<string> PG_sequence_cont;
typedef checked_vector<Specie> PG_species_container;
//Information ueber gefundene Sonden
class PG_inform {
public:
  bool bew;
  int amount; 
  double percentage;
  double quality;
  double qualitysond_sum; // summe aller sonden qualities
  double qualitysond_best; // qualitaet der besten sonde
  double qualitysond_second; // qualitaet der zweitbesten sonde
    
  double fitness;
  PG_species_container treffen_list;
  PG_sequence_cont seq_list;
  PG_inform() { bew = false; }
  void set_bew(bool t){bew=t;}
  double get_quality(){return quality;}
  void set_quality(double q){quality=q;}
  void set_qualitysond_sum(double q){qualitysond_sum=q;}
  void set_qualitysond_best(double q){qualitysond_best=q;}
  void set_qualitysond_second(double q){qualitysond_second=q;}
  void set_amount(int a){amount=a;}
};        


inline void set_test(unsigned x){
  setA seti;
  int a_size = 100;
  while(seti.size()<=x) { 
    int to_insert = rand()%a_size;
    seti.insert(to_insert);
  }
  //	printf("set_test_ende\n");
}

// typedef list<Specie> PG_species_container;
         
class Group {

  PG_species_container       li;
  int size;
  PG_define def;
  PG_inform inf;
  GBDATA *gbmain;
  GBT_TREE *tree;
  const GBT_TREE* lowest_father;
  const PG_tree_nodes& a;
  int birth;
  int group_gen;
  string history;
  bool del_mi;
  int number_leaf;

  int  minprobelength;
  int  maxprobelength;
  double lomintargets;
  double himintargets;
  int lomishit;
  int himishit;

  //koeffizient in Bewertung
  static double ein_sond_wert;
  static double ein_sond_wertq;
    
  static double koefmintargets;
  static double koefmishits;
    
  static double norm;
  static double norm_groesse;

  static bool count_all_probes;
       
public:
 
  Group(GBDATA* gbm, GBT_TREE *tree_in,const PG_tree_nodes& a1);
  //Ein Teil von Gruppe wird zufaellig geaendert
  Group(const Group& gr,double teil,int b);
  //Kopierkonstruktor
  Group(const Group& gr);
  Group& operator=(const Group& other);

  //zwei Gruppe mischen und eine neue erzeugen
  Group(const Group& gr1,double teil1, const Group& gr2,double teil2,int b);
  // zwei Gruppe mischen und zwei neue erzeugen (crossover)
  Group(const Group& gr1,double teil1, const Group& gr2,double teil2,Group& zweite,int b);
  //Species aus Gruppe ausschneiden
  Group(const Group& gr,const PG_species_container& out_l,int b);
  //Species aus Gruppe zufaellig ausschneiden
  Group(const Group& gr,int k,int b);
  //Species zufuegen
  Group(const Group& gr,const PG_species_container& in_l,bool a,int b);
  //Species zufaellig zufuegen
  Group(const Group& gr,int kn,bool abool,int b);
  //Konstruktor
  Group(int x,GBDATA* gbm, GBT_TREE *tree_in,bool v,const PG_tree_nodes& a1,int b,GBDATA*& first_mark_spec);
  //Konstruktor mit Liste
  Group(const PG_species_container& in_l,const PG_tree_nodes& a_in,int b);

  static double get_wert(){return ein_sond_wert;}
  static double get_wertq(){return ein_sond_wertq;}
  static   void  set_wert(double wert){ein_sond_wert=wert;}
  static   void  set_wertq(double wert){ein_sond_wertq=wert;}
  static    void set_koefmintargets(double n){koefmintargets=n;}
  static  double get_koefmintargets(){return koefmintargets;}
  static    void set_koefmishits(double n){koefmishits=n;}
  static   double get_koefmishits(){return koefmishits;}
  static   void set_norm(double n){norm=n;}
  static   void set_norm_groesse(double n){norm_groesse=n;}
  static void set_count_all(bool ca) { count_all_probes = ca; }


  GBT_TREE *ret_tree_pos( GBT_TREE* t,GBDATA *gbspecies);
    
  const PG_inform& get_inf() const { return inf; }
  PG_inform& get_inf() { return inf; }
  void set_inf(const PG_inform& setinform)  { inf=setinform;}

  const PG_define& get_def() const {return def;}
  PG_define& get_def() {return def;}
  void  set_def(const PG_define& setdef){def=setdef;}

  void set_gbmain(GBDATA* gbmainset) {gbmain=gbmainset;}
  GBDATA *get_gbmain() { return gbmain;}
    
  const PG_species_container& get_li() const {return li;} 
  PG_species_container& get_li() {return li;} 

  const PG_tree_nodes& get_a(){return a;}
  int group_size(void);
    
  int get_gen(){return group_gen;}
  void set_gen(int n){group_gen=n;}
  int get_birth(){return birth;}
  const int get_birth() const {return birth;}

  bool get_del_mi(){return del_mi;}
  const bool get_del_mi() const {return del_mi;}
  void set_del_mi(bool b){del_mi=b;}

  GBDATA* return_k_te_specie_indatensatz(int k, GBDATA* gbm);
  const GBT_TREE* return_out_array(int k); 

  PG_species_container::iterator  k_pos(int h);
 
  bool test_doppelte(Specie test,const PG_species_container& l);
  long dist(Specie sp1,Specie sp2);
  long alldist();
  double rel_dist();
                     
  //private:
  bool ist_bewertet(){return inf.bew;}
 
  double  bewertung();
  //{ inf.bew=true;
  //return (inf.ist_sond*100 + inf.anzahl_sond*inf.quality+inf.norm*1/alldist());
  //
  //}
  void toString(); 
  string& get_history(){return history;}
  const string& get_history() const {return history;}
  void set_history(string h){history=h;}  
  void print_history(){printf(" %s\n",history.c_str());} 
  double divid(double f){if (f==0) return 1/2;
  else return 1/f;}

  int cross_probelength(int len1,int len2){return (int)((len1 + len2)/2);}
  int mut_probelength(int i){
    int r=rand()%2;
    int sign=rand()%1;
    if (sign==0) {
      if((i-r)>= minprobelength) return i-r;
      else return i;
    }
    else { /*if (sign==1) */
      if ((i+r)<=maxprobelength) return i+r;
      else return i;
    }
  }
  int cross_mishit(int hit1,int hit2){return (int)((hit1+hit2)/2);}
  int mut_mishit(int i){int r=rand()%2;
  if ((i-r)>=lomishit) return i-r;
  else return i;
  }

  int cross_mintargets(int targ1,int targ2){return (int)((targ1+targ2)/2);}
  int mut_mintargets(int i){int r=rand()%5+1;
  if ((i+r)<=himintargets) return i+r;
  else return i;
  }

  int get_number_leaf(){return number_leaf;}
const int get_number_leaf() const {return number_leaf;}
  void set_number_leaf(int n){number_leaf=n;}

  void set_minprobelength(int d){minprobelength=d;}
  void set_maxprobelength(int d){maxprobelength=d;}
  void set_lomintargets(double d){lomintargets=d;}
  void set_himintargets(double d){himintargets=d;}
  void set_lomishit(int d){lomishit=d;}
  void set_himishit(int d){himishit=d;}
    

  int get_minprobelength(){return minprobelength;}
  int get_maxprobelength(){return maxprobelength;}
  double get_lomintargets(){return lomintargets;}
  double get_himintargets(){return himintargets;}
  int get_lomishit(){return lomishit;}
  int get_himishit(){return himishit;}

  void set_lowest_father(const GBT_TREE* low){lowest_father=low;}
  const GBT_TREE* get_lowest_father(){return lowest_father;}
  const GBT_TREE* get_all_father();
  bool is_father(const GBT_TREE* father,const GBT_TREE* sohn);
  bool is_lowest_father();

};

typedef checked_vector<Group> container_group;

class POOL {

  container_group con;
 container_group match_con;
  GBDATA* gbmain;
  GBT_TREE* tree;
  string PtServerName;
  PG_tree_nodes a;
    
  //Derzeitige Groesse des Pools
  int x;
  //Notwendige Groesse des Pools
  int poolsize;
  //Groesse der Gruppe
  int y;
  double zufall;
  double neue;
  double mischen;
  double treffen;
    
  int  minprobelength;
  int  maxprobelength;
  double lomintargets;
  double himintargets;
  int lomishit;
  int himishit;

  int generation;
  

void Array2(PG_tree_nodes& nodes, GBT_TREE *node);
void  create_markArray(GBT_TREE *tree);
GBT_TREE *ret_tree_pos( GBT_TREE* t,GBDATA *gbspecies);
 
public:
  //Konstruktor fuer Initialisierung
  //x -Anzahl von Gruppen; y- Anzahl Elementen in der Gruppe
  // bool variante:true zufaellig, false als Ausschnitt aus dem
  //Baum
  POOL(GBDATA* gbmain1,GBT_TREE* tree1,const string& ptservername, int x1,int y1,bool variante,bool all_not_all);
  void set_tree(GBT_TREE *in_tree){tree=in_tree;}
  //Sondensuchfunktion durchfueren
  bool search_pool(GBT_TREE* gbt_tree,GB_HASH* hash/*vielleicht noch dazu Parameter*/);
  //Bilden von neuen Gruppen
  bool step(int zufall1,int neue1,int mischen1,int treffen1,int groupsize, PG_inform& pgi,GBT_TREE* gbt_tree_step,GB_HASH* hash_step);
  // void eingeben(int* zufall,int* neue,int* mischen,int* treffen);

  container_group::const_iterator begin() const { return con.begin(); }
  container_group::const_iterator end() const { return con.end(); }
  //Get -Set Funktionen fuer Felder

  const container_group& get_con() const {return con;}
  container_group& get_con() {return con;}

  const container_group& get_match_con() const {return match_con;}
  container_group& get_match_con() {return match_con;}

  const PG_tree_nodes& get_a() const {return a;}
  string get_server_name(){return PtServerName;} 
 
  void set_x(int x1){x=x1;}
  int  get_x(){return x;}

  void  set_y(int y1){y=y1;}
  int get_y(){return y;}

  void set_zufall(double zufall_ein){ zufall=zufall_ein;}
  double get_zufall(){return zufall;}

  void set_neue(double neue_ein){ neue=neue_ein;}
  double get_neue(){return neue;}

  void set_mischen(double mischen_neu){mischen=mischen_neu;}
  double get_mischen(){return mischen;}

  void set_treffen(double treffen_ein){treffen=treffen_ein;}
  double get_treffen(){return treffen;}
    
  void set_minprobelength(int d){minprobelength=d;}
  void set_maxprobelength(int d){maxprobelength=d;}
  void set_lomintargets(double d){lomintargets=d;}
  void set_himintargets(double d){himintargets=d;}
  void set_lomishit(int d){lomishit=d;}
  void set_himishit(int d){himishit=d;}
    

  int get_minprobelength(){return minprobelength;}
  int get_maxprobelength(){return maxprobelength;}
  double get_lomintargets(){return lomintargets;}
  double get_himintargets(){return himintargets;}
  int get_lomishit(){return lomishit;}
  int get_himishit(){return himishit;}

 
  //PG_tree_nodes Array();

  void createArray();
  void replic(int del);

  int cross_probelength(int len1,int len2){return (int)((len1 + len2)/2);}
  int mut_probelength(int i){
    int r=rand()%2;
    int sign=rand()%2;
    if (sign==0) {
      if((i-r)>= minprobelength) return i-r;
      else return i;
    }
    else { /*if (sign==1) */
      if ((i+r)<=maxprobelength) return i+r;
      else return i;
    }
  }
  int cross_mishit(int hit1,int hit2){return (int)((hit1+hit2)/2);}
  int mut_mishit(int i){int r=rand()%4+1;
  if ((i-r)>=lomishit) return i-r;
  else return i;
  }

  int cross_mintargets(int targ1,int targ2){return (int)((targ1+targ2)/2);}
  int mut_mintargets(int i){int r=rand()%5+1;
  if ((i+r)<=himintargets) return i+r;
  else return i;
  }
    
  void  toString() ;
  int get_gen(){return generation;}
 
};

struct less_name : public binary_function<Specie, Specie, bool> {
  bool operator()(Specie x, Specie y) { 
    GB_ERROR err=0;
          
    const  char* namex=x.get_name(err);
    if (err!=0) namex="error";
    const char* namey=y.get_name(err);
    if (err!=0) namey="error";
    return strcmp(namex,namey)<0;
  }
};

class GroupPtr{
  Group* gr;
  bool del_mi;
public:
  int nummer;
  GroupPtr(Group* groupptr,int num){ gr=groupptr; pg_assert(gr); del_mi = false;nummer=num;}
  
  bool get_del_mi() const {return del_mi;}
  void set_del_mi(bool b){del_mi=b;}
  const Group* get_gr() const {return gr;}
  Group* get_gr()  {return gr;}

  bool operator==(const GroupPtr& other) const { return other.get_gr()==get_gr(); }
}
;

struct less_group : public binary_function<GroupPtr, GroupPtr, bool> {
  bool is_less(const GroupPtr& x, const GroupPtr& y) const;
  bool operator()(GroupPtr x, GroupPtr y) const { return is_less(x,y); }
  //   /*
  // (Group x, Group y) {
  //     if (x.get_li().size()<y.get_li.size()) return true;
  //     else if (x.get_li().size()>y.get_li.size()) return false;
    
  //     else


  // }
  // */
};
struct if_del : public unary_function<GroupPtr, bool> {
  bool operator()(GroupPtr x){
    return (x.get_del_mi());
  }
};
typedef /*checked_*/vector<GroupPtr> container_group_ptr;

struct less_group_quality : public binary_function<GroupPtr, GroupPtr, bool> {
  bool operator()(GroupPtr x, GroupPtr y){
    return x.get_gr()->get_inf().quality<y.get_gr()->get_inf().quality;
  }
};


class Tiefe{
  GBT_TREE *tree;
  const GBT_TREE* p;
  int tiefe; 

public:
  Tiefe(GBT_TREE* in,GBT_TREE* wurz){p=in;tree=wurz;tiefe_berechnen();}
Tiefe(const GBT_TREE* cin,GBT_TREE* wurz){p=cin;tree=wurz;tiefe_berechnen();}
  void up(){p=p->father;}
  
  const GBT_TREE* get_tree_pos(){return p;}
  void set_tree_pos( const GBT_TREE* pr){p=pr;tiefe_berechnen();  }
  const GBT_TREE* lowest_father(class Tiefe* other);
  void tiefe_berechnen(); 
  const GBT_TREE* get_p() const {return p;}
  
  //  GBT_TREE* get_p()  {return p;}
  
};

typedef checked_vector<Tiefe> tiefe_container;

#else
#error probe_group2.hxx included twice
#endif // PROBE_GROUP2_HXX

