#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <arbdb.h>

#include <aw_window.hxx>
#include <aw_awars.hxx>
#include <PT_com.h>
#include <client.h>
#include <servercntrl.h>

#include "probe_search.hxx"

//#define AUSGABE
extern GBDATA *gb_main;


static GB_ERROR pd_get_the_names(Group& group, bytestring &bs, bytestring &checksum,
Specie*& bad_species) {

    long	len;
    void *names = GBS_stropen(1024);
    void *checksums = GBS_stropen(1024);

    GB_begin_transaction(gb_main);
    char *use = GBT_get_default_alignment(gb_main);
    GB_ERROR error = 0;

    len = 0;int y=0;;
    for (PG_species_container::const_iterator s=group.get_li().begin(); s!=group.get_li().end(); ++s) {

        const char *spec_name = s->get_name(error);
        y++;
        if (!spec_name) {
            error = GB_export_error("Species has no name");
            bad_species = &*((PG_species_container::iterator)s);
            GB_abort_transaction(gb_main);
            return error;
        }
        const char *spec_data = s->get_aligm(use, error);
        if (!spec_data) {
            error = GB_export_error("Species '%s' has no sequence data", spec_name);
            GB_abort_transaction(gb_main);
            free(use);
            return error;
        }

        GBS_intcat(checksums, GBS_checksum(spec_data, 1, ".-"));
        GBS_strcat(names, spec_name);
        GBS_chrcat(checksums, '#');
        GBS_chrcat(names, '#');
    }

    bs.data = GBS_strclose(names, 0);
    bs.size = strlen(bs.data)+1;

    checksum.data = GBS_strclose(checksums, 0);
    checksum.size = strlen(checksum.data)+1;

    GB_commit_transaction(gb_main);
    free(use);
    return error;
}

static int pg_probe_design_send_data(Group& gr, T_PT_PDC  pdc, struct gl_struct& pd_gl)
{
    //@@@ eventuell nur eimal bei Konstruktion senden

    int i;
    char buffer[256];
    if (aisc_put(pd_gl.link, PT_PDC, pdc,
                 PDC_DTEDGE,gr.get_def().dtedge,
                 PDC_DT,gr.get_def().dt,
                 PDC_SPLIT,gr.get_def().split,
                 PDC_CLIPRESULT, gr.get_def().cliptresult,
                 0)) return 1;
    for (i=0;i<16;i++) {
        sprintf(buffer,"probe_design/bonds/pos%i",i);
        if (aisc_put(pd_gl.link,PT_PDC, pdc,
                     PT_INDEX,	i,
                     PDC_BONDVAL, gr.get_def().bondval[i],
                     0) ) return 1;
    }
    return 0;
}

// --------------------------------------------------------------------------------
//     class PT_server_connection
// --------------------------------------------------------------------------------
class PT_server_connection {
private:
    T_PT_PDC  pdc;
    GB_ERROR error;
    struct gl_struct my_pd_gl;

public:
    PT_server_connection(const char *servername) {
        error = 0;
        memset(&my_pd_gl, 0, sizeof(my_pd_gl));

        char	*servername_1;
        printf("Search a free running server\n");

        servername_1=(char *)probe_pt_look_for_server(servername,error);
        if (servername_1) {
            my_pd_gl.link = (aisc_com *)aisc_open(servername_1, &my_pd_gl.com,AISC_MAGIC_NUMBER);
            free (servername_1); servername_1 = 0;

            if (!my_pd_gl.link) {
                error = GB_export_error("Cannot contact Probe bank server ");
            }
            else if (init_local_com_struct(my_pd_gl)){
                error = "Cannot contact to probe server: Connection Refused";
            }
            else {
                aisc_create(my_pd_gl.link,PT_LOCS, my_pd_gl.locs,
                            LOCS_PROBE_DESIGN_CONFIG, PT_PDC,	&pdc,
                            0);
            }
        }
    }
    virtual ~PT_server_connection() {
        if (my_pd_gl.link) {
            aisc_close(my_pd_gl.link);

        }
        delete error;
    }

    struct gl_struct& get_pd_gl() { return my_pd_gl; }

    GB_ERROR get_error() const { return error; }

    T_PT_PDC& get_pdc() { return pdc; }
};



// --------------------------------------------------------------------------------
//     bool PG_search_probes_and_rate_group(Group& group,const char *servername, Specie *& bad_species)
// --------------------------------------------------------------------------------
bool PG_search_probes_and_rate_group(Group& group,const char *servername, Specie *& bad_species, GB_ERROR& error1)
{


    pg_assert(group.group_size()>0);

      bad_species = 0;
	static PT_server_connection *my_server;



    if (!my_server)  {
        my_server = new PT_server_connection(servername);
        error1 = my_server->get_error();
        if (error1) {
            aw_message(error1);
            return false;
        }
    }

    T_PT_TPROBE tprobe;
    bytestring bs = {0, 0};
    bytestring check = { 0, 0 };
    char	*match_info;


#ifdef AUSGABE
    printf("Probe Design\n");
#endif


    static int count;
    if (!error1){
        count++;
        error1 = pd_get_the_names(group, bs,check, bad_species);
    }

    if (error1){
        aw_message (error1);
        aw_closestatus();
        return false;
    }

    aw_status("Start probe design (Cannot be stopped)");


    aisc_put(my_server->get_pd_gl().link,PT_PDC, my_server->get_pdc(),

             PDC_PROBELENGTH,group.get_def().probelength,
             PDC_MINTEMP,group.get_def().mintemp,
             PDC_MAXTEMP,group.get_def().maxtemp,
             PDC_MINGC,group.get_def().mingc/100.0,
             PDC_MAXGC,group.get_def().maxgc/100.0,
             PDC_MAXBOND,group.get_def().maxbonds,

             PDC_MINPOS,(long)group.get_def().minpos,
             PDC_MAXPOS,(long)group.get_def().maxpos,
             PDC_MISHIT,(long)group.get_def().mishit,
             PDC_MINTARGETS,group.get_def().mintargets/100.0,
             0);
#ifdef AUSGABE
    printf("nach aisc\n");
#endif
    if (pg_probe_design_send_data(group,my_server->get_pdc(),my_server->get_pd_gl())) {
        aw_message ("Connection to PT_SERVER lost (1)");
        aw_closestatus();
 error1 = GB_export_error("Connection to PT_SERVER lost (1)");
	    printf("%s\n",error1);
        return false;
    }


    aisc_put(my_server->get_pd_gl().link,PT_PDC, my_server->get_pdc(),
             PDC_NAMES, &bs,
             PDC_CHECKSUMS, &check,
             0);


    /* Get the unknown names !!! */
    bytestring unknown_names;
    if (aisc_get(my_server->get_pd_gl().link,PT_PDC, my_server->get_pdc(),
                 PDC_UNKNOWN_NAMES, &unknown_names,
                 0)){
      error1 = GB_export_error("Connection to PT_SERVER lost (1)");
	    printf("%s\n",error1);
        aw_closestatus();
        return false;
    }
#ifdef AUSGABE
    printf("unames\n");
#endif
    char *unames = 	unknown_names.data;
    if (unknown_names.size>1) {

        error1 = GB_export_error("Your PT server is not up to date or wrongly chosen\n"
                                     "	The following names are new to it:\n"
                                     "	%s\n"
                                     "	This version allows you to quickly add the unknown sequences\n"
                                     "	to the pt_server\nAborting!\n"
                                 ,unames);

        free(unknown_names.data);
        return false;
    }

    aisc_put(my_server->get_pd_gl().link,PT_PDC, my_server->get_pdc(),
             PDC_GO,0,
             0);

    printf("Read the results from the server\n");
    {
        char *locs_error = 0;
        if (aisc_get( my_server->get_pd_gl().link, PT_LOCS, my_server->get_pd_gl().locs,
                      LOCS_ERROR	,&locs_error,
                      0)){
            error1 = GB_export_error("Connection to PT_SERVER lost (1)");
            printf("%s\n",error1);
            return false;
        }
        if (*locs_error) {
	  error1 = GB_export_error(locs_error);
	  printf("%s\n", locs_error);
	  return false;
        }
        free(locs_error);
    }

    tprobe = 0;
    aisc_get( my_server->get_pd_gl().link, PT_PDC, my_server->get_pdc(),
              PDC_TPROBE, &tprobe,
              0);



    if (tprobe) {
        aisc_get( my_server->get_pd_gl().link, PT_TPROBE, tprobe,
                  TPROBE_INFO_HEADER,	&match_info,
                  0);
        char *s = strtok(match_info,"\n");
        while (s) {
            printf("%s\n", s);
            s = strtok(0,"\n");
        }
        free(match_info); match_info = 0;
    }else{
        printf("There are no results");
    }

    {
        double my_TPROBE_QUALITY;
        char *my_TPROBE_SEQUENCE = 0;
        int my_TPROBE_GROUPSIZE;
        char *my_TPROBE_KEYSTRING = 0;




	group.get_inf().set_amount(0);
	group.get_inf().set_qualitysond_sum(0);
	group.get_inf().set_qualitysond_best(0);
	group.get_inf().set_qualitysond_second(0);
	group.get_inf().set_quality(0);
	group.get_inf().seq_list.erase( group.get_inf().seq_list.begin(), group.get_inf().seq_list.end());

	int i=0;

        while (	tprobe ){
#ifdef AUSGABE
            printf("begin\n");
#endif
            if (aisc_get( my_server->get_pd_gl().link, PT_TPROBE, tprobe,
                          TPROBE_NEXT,		&tprobe,
                          TPROBE_INFO,		&match_info,
                          TPROBE_GROUPSIZE, &my_TPROBE_GROUPSIZE, // groesse der Gruppe,  die von Sonde getroffen wird?
                          TPROBE_KEYSTRING, &my_TPROBE_KEYSTRING,
                          TPROBE_SEQUENCE, &my_TPROBE_SEQUENCE,  // Sondensequenz codiert (2=A 3=C 4=G 5=U)
                          TPROBE_QUALITY, &my_TPROBE_QUALITY, // Qualitaet der Sonde ?
                          0)) break;

            pg_assert(my_TPROBE_SEQUENCE[0]);

            for(int j=0;my_TPROBE_SEQUENCE[j];++j) {
                switch (my_TPROBE_SEQUENCE[j]) {
                case 2: my_TPROBE_SEQUENCE[j] = 'A'; break;
                case 3: my_TPROBE_SEQUENCE[j] = 'C'; break;
                case 4: my_TPROBE_SEQUENCE[j] = 'G'; break;
                case 5: my_TPROBE_SEQUENCE[j] = 'U'; break;
                default:
                    pg_assert(0);
                    break;
                }
            }
#ifdef AUSGABE
            printf(" TPROBE_SEQUENCE=%s\n",my_TPROBE_SEQUENCE);
#endif
            i++;
#ifdef AUSGABE
printf("i=%d\n",i);
#endif
            group.get_inf().set_amount(group.get_inf().amount+1);
            group.get_inf().qualitysond_sum=
	      group.get_inf().qualitysond_sum + my_TPROBE_QUALITY;


	    if (my_TPROBE_QUALITY > group.get_inf().qualitysond_second) {
	      if (my_TPROBE_QUALITY > group.get_inf().qualitysond_best) {
		group.get_inf().qualitysond_second = group.get_inf().qualitysond_best;
		group.get_inf().qualitysond_best = my_TPROBE_QUALITY;
	      }
	      else {
		group.get_inf().qualitysond_second = my_TPROBE_QUALITY;
	      }
	    }

            group.get_inf().seq_list.push_back(my_TPROBE_SEQUENCE);
#ifdef AUSGABE
            printf("free\n");
#endif
            free(my_TPROBE_KEYSTRING); my_TPROBE_KEYSTRING = 0;
            free(my_TPROBE_SEQUENCE); my_TPROBE_SEQUENCE = 0;
            free(match_info); match_info = 0;
        }


    }

#ifdef AUSGABE
    printf("vor_end\n");
#endif


    delete bs.data;
    delete check.data;


#ifdef AUSGABE
    printf("PG_search_probes_and_rate_group\n");
#endif
    return true;
}



void probe_match_event(Group& group,const char *servername, Specie *& bad_species, GB_ERROR& error1,GBT_TREE* gbt_tree,GB_HASH* hash
)
{

//     char	result[1024];

    T_PT_MATCHLIST match_list;
    char *match_info, *match_name;
    //int	mark;
    const char	*probe;
    char	*locs_error;

    GBDATA *gb_species_data = 0;
    GBDATA *gb_species;
    int show_status = 0;
    //int extras = 1; // mark species, write to temp fields,

	static PT_server_connection *my_server;
	GBT_TREE* p=0;



 if (!my_server)  {
        my_server = new PT_server_connection(servername);
        error1 = my_server->get_error();
        if (error1) {
            aw_message(error1);
            return ;
        }
    }


    if (pg_probe_design_send_data(group,my_server->get_pdc(),my_server->get_pd_gl())) {
        aw_message ("Connection to PT_SERVER lost (2)");

        printf("Connection to PT_SERVER lost (2)\n");
        return;
    }
    if (group.get_inf().seq_list.size()>0)
	  probe =group.get_inf().seq_list[0].c_str();
	else {printf("Keine Sonden \n");exit(0);}



// printf("Start Probe Match\n");

    if (aisc_nput(my_server->get_pd_gl().link,PT_LOCS, my_server->get_pd_gl().locs,
                  LOCS_MATCH_REVERSED,	0	,
                  LOCS_MATCH_SORT_BY,	0 ,
                  LOCS_MATCH_COMPLEMENT,		0,
                  LOCS_MATCH_MAX_MISMATCHES,0,
                  LOCS_MATCH_MAX_SPECIES, 200,
                  LOCS_SEARCHMATCH,		probe,
                  0)){

        aw_message ("Connection to PT_SERVER lost (2)");

        return;
    }


         GB_begin_transaction(gb_main);
         gb_species_data = GB_search(gb_main,"species_data",GB_CREATE_CONTAINER);


    long match_list_cnt;
    bytestring bs;
    bs.data = 0;

// printf("Read the Results\n");
    if (aisc_get( my_server->get_pd_gl().link, PT_LOCS, my_server->get_pd_gl().locs,
                  LOCS_MATCH_LIST,	&match_list,
                  LOCS_MATCH_LIST_CNT,	&match_list_cnt,
                  LOCS_MATCH_STRING,	&bs,
                  LOCS_ERROR,		&locs_error,
                  0)){
        aw_message ("Connection to PT_SERVER lost (3)");
        if (show_status) aw_closestatus();
    }

    if (*locs_error) {
        aw_message(locs_error);
    }
    free(locs_error);



    //long mcount = 0;
    char toksep[2];
    toksep[0] = 1;
    toksep[1] = 0;
    char *hinfo = strtok(bs.data,toksep);



    while (hinfo && (match_name = strtok(0,toksep)) ) {

        match_info = strtok(0,toksep);
        if (!match_info) break;

        if (gb_main){
            gb_species = GBT_find_species_rel_species_data(gb_species_data,match_name); // species zeiger suchen

            //
            // GBDATA *gb_full_name = GB_search(gb_species, "full_name", GB_STRING);
            // char *xxx            = GB_read_string(gb_full_name);
            // const char *xxx      = GB_read_char_pntr(gb_full_name);


	    p=(GBT_TREE*)GBS_read_hash(hash,match_name);
	    if (p!=0)
	    {
		class Specie sp( p) ;
		if ( !(group.test_doppelte(sp,group.get_li())) )
		    group.get_li().push_back(sp);
	    }
	    else {

	    }

        }


    }



    delete bs.data;
    if (gb_main) GB_commit_transaction(gb_main);



 group.get_inf(). bew=false;
 group.get_def().mishit=0;


    return;
}

